import os
import traceback
import rasterio
import rasterio.features
import rasterio.warp
import shapely.geometry
from geojson import Feature
from osgeo import osr, gdal
from starearth.coordinate import EPSG
from starearth.utils.mer_geo import NewGlobalMercator


def create_default_geojsonl(input_file, geojsonl, properties=None):
    """
    根据输入的tif数据，生成同名的geojsonl
    :param input_file: tif文件的绝对路径
    :param geojsonl: geojsonl文件的绝对路径
    :param properties: 各种属性
    :return: None
    """

    if not properties:
        properties = {"item_id": "default", "date": "default"}

        pass

    # ds = gdal.Open(new_input_file)
    # temp = '{"type": "Feature","properties": {}, "geometry": {}}'
    # [[[],[],[],[],[]]]
    # gt = ds.GetGeoTransform()
    # width, height = ds.RasterXSize, ds.RasterYSize
    # left, top = gt[0], gt[3]
    # res_x, res_y = gt[1], gt[5]
    # right, bottom = left + res_x * width, top + res_y * height
    # 计算真实角点坐标（算法来自李志明），TODO：历史遗留 不规则影像需要给定影像的几何范围：shape file或geojson,否则计算不准确
    # row_top, row_bottom, column_left, column_right = getRealFourCorners(new_input_file)
    # print('角点坐标：%s,%s,%s,%s' % (row_top, row_bottom, column_left, column_right))
    # # left, top = row_top
    # # right, bottom = row_bottom
    # if epsg != 'EPSG:4326':
    #     from_srs_id = int(epsg.split(':')[-1])
    #     row_top = transform_point(point=row_top, from_srs_id=from_srs_id, to_srs_id=4326)
    #     row_bottom = transform_point(point=row_bottom, from_srs_id=from_srs_id, to_srs_id=4326)
    #     column_left = transform_point(point=column_left, from_srs_id=from_srs_id, to_srs_id=4326)
    #     column_right = transform_point(point=column_right, from_srs_id=from_srs_id, to_srs_id=4326)
    #
    # coordinates = [
    #     [list(column_left[:2]), list(row_top[:2]), list(column_right[:2]), list(row_bottom[:2]), list(column_left[:2])]]
    # temp = """{{"type": "Feature","properties": {"item_id": "default","date": "default"},"geometry": {}}}"""

    # 判断是否有投影信息，没有则赋值一个默认的4326
    ds = gdal.Open(input_file)
    crs = ds.GetProjection()

    if not crs:
        default_wkt = EPSG(4326).wkt
        ds.SetProjection(default_wkt)

    # rasterio不支持pix格式数据，以及libjpeg: Unsupported JPEG data precision 12，可能还有其他数据打开错误
    # 对于rasterio无法处理的数据，统一转换成tif数据后，再进行处理
    try:

        out_tif_path = ''

        with rasterio.open(input_file, 'r') as dataset:
            dataset.dataset_mask()

    except Exception as err:

        basepath = os.path.dirname(input_file)
        basename = os.path.basename(input_file)
        filename = basename.split('.')[0] + '.tif'

        out_tif_path = os.path.join(basepath, filename)
        input_file = out_tif_path  # 根据转换后的tif数据来生成geojsonl

        ds = gdal.Translate(out_tif_path, ds)

        pass

    del ds

    # 新版算法
    geojsonl_content = """"""

    with rasterio.open(input_file, 'r') as dataset:

        mask = dataset.dataset_mask()

        for polygon, value in rasterio.features.shapes(mask, transform=dataset.transform):
            # geometry: {'type': 'Polygon', 'coordinates': [[[36.819799, 23.443863],[36.911464, 23.385531],[36.911464, 23.335532],[36.919797, 23.335532],[36.819799, 23.443863]]]}
            # value: 255 or 0
            geometry = rasterio.warp.transform_geom \
                    (
                    dataset.crs,
                    "EPSG:4326",
                    polygon,
                    precision=6
                )

            # 对geojson范围进行化简
            origin_geom = shapely.geometry.shape(geometry)
            simplified_geom = origin_geom.simplify(0.0001)

            feature = Feature(type="Feature", properties=properties, geometry=simplified_geom)

            geojsonl_content = geojsonl_content + str(feature) + '\n'

    with open(str(geojsonl), 'w') as f:

        f.write(geojsonl_content)

    # 删除转换后的tif数据
    if out_tif_path:
        os.remove(out_tif_path)

        pass

    pass


def calc_minmax_zoom(new_input_file, epsg, image_tile_size):
    '''
    自动计算切片级别
    :param new_input_file: 输入的影像文件
    :param epsg: 输入影像的投影，形如`EPSG:3857`
    :return: 最小级别min_z, 最大级别max_z
    '''
    ds = gdal.Open(new_input_file)
    if epsg != 'EPSG:3857':
        from_srs = osr.SpatialReference()
        from_srs.ImportFromEPSG(int(epsg.split(':')[1]))
        to_srs = osr.SpatialReference()
        to_srs.ImportFromEPSG(3857)
        ds = gdal.AutoCreateWarpedVRT(ds, from_srs.ExportToWkt(), to_srs.ExportToWkt())

    gt = ds.GetGeoTransform()

    mercator = NewGlobalMercator(tile_size=image_tile_size)

    # 根据影像的大小和瓦片大小确定最小切片级别
    min_z = mercator.ZoomForPixelSize(gt[1] * max(ds.RasterXSize, ds.RasterYSize) / float(image_tile_size))

    # 根据分辨率确定最大切片级别
    max_z = mercator.ZoomForPixelSize(gt[1])

    return min_z, max_z


def calc_minmax_zoom2(input_file, epsg, image_tile_size):
    """
    自动计算切片级别2
    :param new_input_file: 输入的影像文件
    :param epsg: 输入影像的投影，形如`EPSG:3857`
    :return: 最小级别min_z, 最大级别max_z
    """
    ds = gdal.Open(input_file)
    if epsg != 'EPSG:3857':
        if isinstance(epsg, int):
            epsg_id = epsg
        else:  # "EPSG:4326"
            epsg_id = int(epsg.split(':')[1])
        from_srs = osr.SpatialReference()
        from_srs.ImportFromEPSG(epsg_id)
        to_srs = osr.SpatialReference()
        to_srs.ImportFromEPSG(3857)
        ds = gdal.AutoCreateWarpedVRT(ds, from_srs.ExportToWkt(), to_srs.ExportToWkt())

    gt = ds.GetGeoTransform()

    # mercator = NewGlobalMercator(tile_size=IMAGE_TILE_SIZE)
    mercator = NewGlobalMercator(tile_size=image_tile_size)
    # 根据分辨率确定最大切片级别
    max_z = mercator.ZoomForPixelSize(gt[1])
    # 最小切片级别设为最大切片级别-3，即自动切4级
    min_z = max_z - 3

    return min_z, max_z


def verify_geotiff(input_file, epsg):
    """
    @param: new_input_file, 输入文件路径
    @param: epsg, 参数可能为: None, '', int或不合法的字符串
    验证geotiff，传入epsg字符串，返回经过验证的epsg字符串或抛出异常，形如"EPSG:4326"
    1. 当epsg是int值或'EPSG:4326'字符串时，程序先得到epsg_id
    2. 获取原始文件内置投影信息，
              如与输入参数不一致报错；相同继续；
              如原始文件无投影信息，会先添加投影信息落盘，再返回epsg
    3. 当传入的epsg为None或''时，从原始文件中获取投影，如无投影信息则报错；如获取到投影时则继续；
    """
    # 启用gdal底层异常捕获
    gdal.UseExceptions()

    try:
        ds = gdal.Open(input_file, gdal.GA_Update)
    except Exception as err:
        ds = gdal.Open(input_file)  # 打开jpg和png格式的数据

    if ds is None:  # 如果tif没有内置投影信息，同时没有配套的tfw数据，会发生这种情况。
        raise Exception('数据未能正常打开或不支持给定的数据格式.')

    origin_wkt = ds.GetProjection()

    if epsg:

        if isinstance(epsg, int):
            epsg_id = epsg
            epsg = 'EPSG:%s' % epsg_id

        else:  # "EPSG:4326"
            try:
                epsg_id = int(epsg.split(':')[1])

            except AttributeError:
                raise Exception('epsg字符串格式错误，正确的格式类似：EPSG:4326.')

            except ValueError:
                raise Exception('epsg字符串格式错误，正确的格式类似：EPSG:4326.')

        input_srs = osr.SpatialReference()
        input_srs.ImportFromEPSG(epsg_id)

        if origin_wkt:

            origin_srs = osr.SpatialReference()
            origin_srs.ImportFromWkt(origin_wkt)

            if input_srs.ExportToProj4().strip() != origin_srs.ExportToProj4().strip():
                raise Exception('指定的数据坐标系统与原始坐标坐标系统不一致.')

        else:

            sr = osr.SpatialReference()
            sr.ImportFromEPSG(epsg_id)
            wkt = sr.ExportToWkt()
            ds.SetProjection(wkt)
            # 落盘为安，以便下次读取时能够获取正确的信息。
            ds.FlushCache()

    else:

        if not origin_wkt:
            raise Exception('数据不合法：没有投影信息')

        origin_srs = osr.SpatialReference()
        origin_srs.ImportFromWkt(origin_wkt)

        if origin_srs.IsProjected():
            epsg_id = origin_srs.GetAuthorityCode(str("PROJCS"))

        else:
            epsg_id = origin_srs.GetAuthorityCode(str("GEOGCS"))

        if epsg_id:
            epsg = 'EPSG:%s' % epsg_id

        else:
            raise Exception('获取不到epsg')

    # 关闭gdal底层异常捕获
    gdal.DontUseExceptions()

    return epsg
