import math
import os
from osgeo import gdal, osr

from .bound      import Bound, V2
from .coordinate import Coordinate
from .datatype   import DataType
from .filename   import Filename


class Scene:

    _bound:      Bound
    _coordinate: Coordinate
    _datatype:   DataType

    def __init__(self):

        self.filename = None

        self._dataset = None
        self._wkt     = None
        self._bound      = None
        self._coordinate = None
        self._resolution = None

        self._interpretation = None

        self._datatype = None
        self._nodata   = None

        self._transform = None

        pass

    @property
    def coordinate(self):

        return self._coordinate

    @property
    def nodata(self):

        return self._nodata

    @property
    def interpretation(self):

        return self._interpretation

    @property
    def dataset(self):

        return self._dataset

    @property
    def transform(self):

        return self._transform

    @property
    def projection(self):

        return self._coordinate

    @property
    def bound(self) -> Bound:

        return self._bound

    @property
    def resolution(self):

        return self._resolution

    @property
    def width(self) -> int:

        return self._dataset.RasterXSize

    @property
    def height(self) -> int:

        return self._dataset.RasterYSize

    @property
    def bands(self) -> int:

        return self._dataset.RasterCount

    @property
    def datatype(self):

        return self._datatype

    @staticmethod
    def information(filename):

        if isinstance(filename, str):

            filename = Filename(filename)

            pass

        dataset: gdal.Dataset = gdal.Open(str(filename), gdal.GA_ReadOnly)

        if dataset is None:

            raise Exception(f"{filename} open failed.")

        return gdal.Info \
            (
                dataset,
                options=gdal.InfoOptions(format="json")
            )

    def load(self, filename):

        if isinstance(filename, str):

            filename = Filename(filename)

            pass

        dataset: gdal.Dataset = gdal.Open(str(filename), gdal.GA_ReadOnly)

        if dataset is None:

            raise Exception(f"{filename} open failed.")

        # 优先使用参数指定的投影信息
        # 如果参数没有指定投影，那么使用自带的
        # 没有自带投影，那么使用一个默认值

        self._coordinate = Coordinate.from_wkt(dataset.GetProjection())

        self._init(dataset)

        self.filename = filename

        self._dataset = dataset

        pass

    def project(self, coordinate):

        if coordinate == self._coordinate:

            return

        dataset = self._dataset

        dataset = Coordinate.warp(dataset, coordinate)

        self._coordinate = coordinate

        self._init(dataset)

        self._dataset = dataset

        pass

    def build_overviews(self):

        self._dataset.BuildOverviews("AVERAGE", [2, 4, 8, 16, 32])

        pass

    def sample(self, bound: Bound, width: int, height: int, span: tuple[int, int] = None):

        """
        在场景中重采样并返回结果。

        :param bound:  经纬度边界
        :param width:  期望输出的像素宽度
        :param height: 期望输出的像素高度
        :param span:   输出像素扩展尺寸（目前仅满足来自 DEM 切片需求，128x128 向右下方扩展至 129x129）
        :return:       采样后的波段数据以及其相对瓦片坐标的位置信息；
                       当边界与场景没有交集时，返回 None。
        """

        # 计算左下角位置

        tx = (bound.left   - self.bound.left  ) / self.resolution.x
        ty = (bound.bottom - self.bound.bottom) / self.resolution.y

        # 计算全图尺寸

        tw = bound.width  / self.resolution.x
        th = bound.height / self.resolution.y

        # 图像和瓦片的地理信息都是右手坐标系，像素操作需要切换至左手坐标系
        # bi是输入的tif数据，bt是想挖取的bound，bi只有部分和bt重叠，那么就只挖取这部分

        bi = Bound.LH_XYWH(0.0,                   0.0, float(self.width), float(self.height))
        bt = Bound.LH_XYWH( tx, self.height - th - ty,                tw,                 th)  # 反转 Y 轴

        # 相交

        bp = bi.intersect(bt)

        if bp is None:

            return None

        # 映射至输出

        rx = bt.width  / width
        ry = bt.height / height

        x = int(bp.x)
        y = int(bp.y)

        ox = int(-bt.x / rx) if bt.x < 0.0 else 0
        oy = int(-bt.y / ry) if bt.y < 0.0 else 0

        if span:

            width  += span[0]
            height += span[1]

            w = int(bp.width  + span[0] * rx)
            h = int(bp.height + span[1] * ry)

            ow = int(math.ceil(bp.width  / rx)) + span[0]
            oh = int(math.ceil(bp.height / ry)) + span[1]

            pass

        else:

            w = int(bp.width )
            h = int(bp.height)

            ow = int(math.ceil(bp.width  / rx))
            oh = int(math.ceil(bp.height / ry))

            pass

        # 越界检查

        w = (self.width  - x) if (x + w) > self.width  else w
        h = (self.height - y) if (y + h) > self.height else h

        ox = (width  - ow) if (ox + ow) > width  else ox
        oy = (height - oh) if (oy + oh) > height else oy

        # 重采样

        data = self._dataset.ReadAsArray(x, y, w, h, None, ow, oh)

        return \
            {
                'data': data,
                'datatype': self._datatype,
                'width': width,
                'height': height,
                'ox': ox,
                'oy': oy,
                'ow': ow,
                'oh': oh
            }

    def _init(self, dataset):

        info = gdal.Info \
                (
                dataset,
                options=gdal.InfoOptions(format="json")
            )

        width  = dataset.RasterXSize
        height = dataset.RasterYSize

        self._transform = dataset.GetGeoTransform()

        self._bound = Bound.from_transform(width, height, self._transform)

        self._resolution = V2(self._transform[1], -self._transform[5])

        band = dataset.GetRasterBand(1)

        bands = info["bands"]

        if band.DataType == gdal.GDT_Byte:

            min_max = (0, 255)

        elif "min" in bands[0] and "max" in bands[0]:

            min_max = (bands[0]["min"], bands[0]["max"])

            pass

        else:

            min_max = band.ComputeRasterMinMax()

            pass

        self._datatype = DataType.from_gdal(band.DataType, min_max)

        self._interpretation = \
            [
                {
                    "band": band,
                    "type": self.datatype.name,
                    "colorInterpretation": gdal.GetColorInterpretationName(dataset.GetRasterBand(band).GetColorInterpretation()),
                    "noDataValue": dataset.GetRasterBand(band).GetNoDataValue()
                }
                for band in range(1, dataset.RasterCount + 1)
            ]

        self._nodata = \
            [
                dataset.GetRasterBand(index).GetNoDataValue()
                for index in range(1, dataset.RasterCount + 1)
            ]

        pass

    pass
