import copy
import json
import os
import shapefile
import shapely.geometry
from geojson import Feature
from datetime import datetime, date
from osgeo import gdal
from app.defines import MISSION_TYPE, ONE_E_MINUS4, ONE_E_MINUS6


class ComplexEncoder(json.JSONEncoder):

    def default(self, obj):

        if isinstance(obj, datetime):
            return obj.strftime('%Y-%m-%d %H:%M:%S')

        elif isinstance(obj, date):
            return obj.strftime("%Y-%m-%d")

        else:
            return json.JSONEncoder.default(self, obj)


def convert_date_time(date_time: str):
    """将datetime转换成%Y-%m-%d %H:%M%S的格式
    将date转换成%Y-%m-%d的格式
    """
    date_time_len = len(date_time)

    # "2014-02-15"
    if date_time_len == 10:
        _date = date_time
        date_time = "{} 00:00:00".format(date_time)

    # "2014-02-15 00:00:00"
    elif date_time_len == 19:
        _date = date_time.split(' ')[0]

    else:
        _date = date_time

    return _date, date_time


def shp_to_geojsonl(field_list, filesystem):
    
    sf_type_mapping = {
        3: 'LineString',
        5: 'Polygon',
    }

    # 将用户手动输入的信息变成 k:v，保存到geojsonl的properties中
    for field_list_one in field_list:

        path_list = field_list_one['path_list']
        field_dict = field_list_one['field_dict']

        for k, v in field_dict.items():

            if not isinstance(v, dict):
                continue

            # type是constant，表明是该字段的值用户手工输入的
            # type是field，表明该字段的值是用户指定shp数据某个字段
            if v['type'] == 'constant':
                field_dict[k] = v['value']

            pass

        for shp_relative_path in path_list:
            filename_shp = os.path.basename(shp_relative_path)
            filename, ext = os.path.splitext(filename_shp)

            # 拼接shp文件路径，以及geojsonl
            shp_file_path = str(filesystem.path(shp_relative_path))
            geojsonl_file_path = shp_file_path.replace('.shp', '.geojsonl')

            # 读取shp文件
            sf = shapefile.Reader(shp_file_path)
            sf_type = sf_type_mapping[sf.shapeType]

            borders = sf.shapes()
            shp_fields = sf.fields
            shp_records = sf.records()

            # 记录shp数据所有字段的下标，key是字段名，value是下标
            shp_fields_index_dic = {}

            minus_num = 0

            for index, field in enumerate(shp_fields):

                if isinstance(field, tuple):
                    minus_num += 1
                    continue

                shp_fields_index_dic[str(field[0])] = index - minus_num

            all_features = []

            # 判断是否有与shp文件同名的tif，如果没有则无法继续
            abs_path = os.path.dirname(shp_file_path)

            shpfilename = os.path.basename(shp_file_path)
            name, ext = os.path.splitext(shpfilename)

            # 查找shp同名的tif
            for _file in os.listdir(abs_path):

                tif_path = os.path.join(abs_path, _file)

                if not os.path.isfile(tif_path):
                    continue

                _name, _ext = os.path.splitext(_file)

                if _ext.lower() not in ['.tif', '.tiff', ] or _name != name:
                    continue

                break

            ds = gdal.Open(tif_path)
            res = float(ds.GetGeoTransform()[1])

            # 根据用户指定的字段名，从shp数据中获取值，并转换成 k:v，保存到geojsonl的properties中
            for record_index, record in enumerate(shp_records):

                properties = copy.deepcopy(field_dict)

                for k, v in properties.items():

                    if not isinstance(v, dict):
                        continue

                    # 根据用户指定的字段名获取该字段在shp数据中的位置，以取得当前要素该字段的值
                    properties[k] = record[shp_fields_index_dic[v['value']]]

                # 处理datetime、date、item_id三个字段
                _date, date_time = convert_date_time(str(properties['datetime']).strip())
                properties['date'] = _date
                properties['datetime'] = date_time
                properties['item_id'] = filename

                # 根据用户提供的数据样例，李老师规定 off_nadir、sun_elev的字段值存入bk1，分号隔开，accuracy的字段值存入bk2，cloudcover的字段值存入bk3
                off_nadir_index = shp_fields_index_dic.get('off_nadir', '')
                sun_elev_index = shp_fields_index_dic.get('sun_elev', '')
                accuracy_index = shp_fields_index_dic.get('accuracy', '')
                cloudcover_index = shp_fields_index_dic.get('cloudcover', '')

                bk1_data = ''
                bk2_data = ''
                bk3_data = ''

                if isinstance(off_nadir_index, int):
                    bk1_data += '{}:'.format(record[off_nadir_index])

                if isinstance(sun_elev_index, int):
                    bk1_data += ':{}'.format(record[sun_elev_index])
                    bk1_data = bk1_data.replace('::', ':')

                if isinstance(accuracy_index, int):
                    bk2_data = record[accuracy_index]

                if isinstance(cloudcover_index, int):
                    bk3_data = record[cloudcover_index]

                properties['bk1'] = bk1_data
                properties['bk2'] = bk2_data
                properties['bk3'] = bk3_data

                # if 'disp' not in properties:
                properties['disp'] = 1

                # if 'res' not in properties:
                properties['res'] = res

                coordinates = [list(li) for li in borders[record_index].points]

                geometry = {
                    'type': sf_type,
                    'coordinates': [coordinates, ],
                }

                # 调用simplify函数进行几何简化
                origin_geom = shapely.geometry.shape(geometry)  # 从GeoJSON得到shapely的geometry对象
                simplified_geom = origin_geom.simplify(ONE_E_MINUS4)

                # geojson转feature
                feature = Feature(geometry=simplified_geom)
                feature["properties"] = properties

                all_features.append(feature)

            # 存储当前的geojsonl，所有的要素记录
            with open(geojsonl_file_path, 'w') as fb:

                for feature in all_features:

                    json.dump(feature, fb, cls=ComplexEncoder)
                    fb.write('\n')

    pass
