"""

使用 FastDFS 进行文件的存储/读取

目前的 fdfs_client 限制：

1. tracker / storage 无法运行于同一个 docker compose 环境
2. 没有文件系统概念，只能依赖第三方（Redis）的键值映射模拟文件系统
3. 自带 metadata 记录机制需要进行二次通讯，不方便也不安全

"""


import json
import os

from enum import Enum

import redis

from fdfs_client.client import get_tracker_conf, Fdfs_client


class FSNode:

    class Type(Enum):

        Directory = 1
        File      = 2

        pass

    def __init__(self, type, name):

        self._type     = type
        self._name     = name
        self._children = {}

        pass

    def __contains__(self, name):

        return name in self._children

    def __str__(self):

        return "\n".join(self._to_string_lines(""))

    @property
    def type(self):

        return self._type

    @property
    def name(self):

        return self._name

    @property
    def children(self):

        return self._children

    def add(self, node):

        self._children[FSNode._key(node)] = node

        return node

    def get(self, name):

        return self._children.get(name)

    def _to_string_lines(self, indent=""):

        data = [f"{indent}{self.name}"]

        indent += "  "

        if self._type == FSNode.Type.Directory:

            directories = []
            files       = []

            for name, node in self._children.items():

                if node.type == FSNode.Type.Directory:

                    directories.append(name)

                    pass

                else:

                    files.append(name)

                    pass

                pass

            directories = sorted(directories)
            files       = sorted(files)

            for name in directories:

                node: FSNode = self._children[name]

                data.extend(node._to_string_lines(indent))

                pass

            for name in files:

                node: FSNode = self._children[name]

                data.extend(node._to_string_lines(indent))

                pass

            pass

        return data

    @staticmethod
    def _key(node):

        return f"{node.name}"

    pass


class FastDFS:

    class Error(Exception):

        pass

    def __init__(self, host, port, db):

        pool = redis.ConnectionPool(host=host, port=port, db=db)

        self._redis = redis.Redis(connection_pool=pool)

        tracker = get_tracker_conf("/work/starearth_tiler/conf/fastdfs.conf")

        self._client = Fdfs_client(tracker)

        pass

    def put(self, file_local: str, file_remote: str, metadata=None):

        """

        将本地文件存储至远程

        :param file_local:  本地文件名（可包含路径）
        :param file_remote: 远程文件名（可包含路径）
        :param metadata:    可选附带元数据（字典对象）

        """

        # 检查本地文件

        FastDFS._check_local_file(file_local)

        # 确保目标路径文件名格式正确

        FastDFS._verify_remote_name(file_remote)

        # 分解路径文件名

        directories, name = FastDFS._split(file_remote)

        if name is None:

            name = os.path.basename(file_local)

            pass

        if metadata is None:

            metadata = {}

            pass

        # 总是上传新文件

        info = self._upload(file_local)

        info     = json.dumps(info,     ensure_ascii=False)
        metadata = json.dumps(metadata, ensure_ascii=False)

        key = self._key(directories, name)

        # 如果存在旧文件索引，删除文件

        value = self._redis.get(key)

        if value:

            value = value.decode()

            file_id = json.loads(value.split("$")[0])["id"].encode()

            self._client.delete_file(file_id)

            pass

        # 记录（覆盖）新文件索引

        value = f"{info}${metadata}"

        self._redis.set(key, value)

        pass

    def get(self, file_remote: str, file_local: str):

        """

        将远程文件获取至本地

        :param file_remote: 远程文件名（可包含路径）
        :param file_local:  本地文件名（可包含路径）

        :return: 文件附带的元数据（字典对象）

        """

        # 确保目标路径文件名格式正确

        FastDFS._verify_local_name(file_remote)

        # 分解路径文件名

        directories, name = FastDFS._split(file_remote)

        if name is None:

            raise FastDFS.Error(f"{file_remote} must be a file")

        # 检查文件索引

        key = self._key(directories, name)

        value = self._redis.get(key)

        if not value:

            raise FastDFS.Error(f"{file_remote} not exist")

        # 确保目标路径文件名格式正确

        FastDFS._verify_local_name(file_local)

        # 如果需要，创建必要的目录

        FastDFS._ensure_path_for_local_file(file_local)

        # 下载文件

        value = value.decode().split("$")

        info     = json.loads(value[0])
        metadata = json.loads(value[1])

        file_id = info["id"].encode()

        self._client.download_to_file(file_local, file_id)

        return metadata

    def tree(self):

        root = FSNode(FSNode.Type.Directory, "/")

        cursor = 0

        while True:

            cursor, records = self._redis.scan(cursor, f"{FastDFS._redis_prefix}*")

            for data in records:

                node = root

                data = data.decode()[14:].split(":")  # 14 == len(FastDFS._redis_prefix)

                directories = data[:-1]
                name        = data[-1]

                for directory in directories:

                    if directory not in node:

                        node = node.add(FSNode(FSNode.Type.Directory, directory))

                        pass

                    else:

                        node = node.get(directory)

                        pass

                    pass

                node.add(FSNode(FSNode.Type.File, name))

                pass

            if cursor == 0:

                break
                pass

            pass

        return root

    _redis_prefix = "starearth:dfs"

    @staticmethod
    def _key(directories, name):

        return f"{FastDFS._redis_prefix}:{':'.join(directories)}:{name}" if directories else name

    @staticmethod
    def _check_local_file(file: str):

        # 必须存在

        if not os.path.exists(file):

            raise FastDFS.Error(f"{file} not exist")

        # 不能是目录

        if os.path.isdir(file):

            raise FastDFS.Error(f"{file} must be a file")

        pass

    @staticmethod
    def _verify_local_name(file: str):

        # todo 校验本地路径文件名是否合法，可能涉及多平台
        # raise
        pass

    @staticmethod
    def _verify_remote_name(file: str):

        # todo 校验远程路径文件名是否合法，仅使用 UNIX 路径文件风格
        # raise
        pass

    @staticmethod
    def _ensure_path_for_local_file(file: str):

        # todo 如果 file_local 中路径不存在，需要创建
        # raise
        pass

    @staticmethod
    def _split(file: str):
        """

        将文件字符串分解为目录数组和文件名

        :param file: 文件字符串

        :return: 目录数组，文件名

        """

        # 规范首斜杠

        if not file.startswith("/"):

            file = f"/{file}"

            pass

        array = file.split("/")[1:]

        directories = array[:-1]
        name        = None if array[-1] == "" else array[-1]

        return directories, name

    def _upload(self, file: str):

        try:

            result = self._client.upload_by_filename(file)

            info = \
                {
                    "group": result["Group name"    ].decode(),
                    "id":    result["Remote file_id"].decode(),
                    "size":  result["Uploaded size" ]
                }

            return info

        except Exception as error:

            raise FastDFS.Error(f"{file} uploading error: {error}")

        pass

    pass


class StorageFDFS:

    # todo
    #  目前可以使用 FastDFS 进行通用文件的操作
    #  在此基础上，使用更早之前写的文件遍历逻辑
    #  将瓦片结果目录下的数据放至 FastDFS
    #  如果需要传入来自 config 的配置，自行修改 FastDFS 构造函数
    #  注意：FastDFS 使用的 Redis 和任务管理使用的 Redis 不是同一个

    # 示例

    """
        try:

            fs = FastDFS()

            fs.put("./aaa/bbb/default/xyz.ptp", "/some/dir/xyz1.ptp", { "可以改名": "可以不改名" })
            fs.put("./aaa/bbb/default/xyz2.ptp", "/some/dir/", { "可以只使用路径，自动使用原名": "但是后面必须跟斜杠，不然会被认为是文件名" })

            metadata = fs.get("some/dir/xyz3.ptp", "./xyz2.ptp", { "同名文件": "会被覆盖" })

            print(metadata)  # { "同名文件": "会被覆盖" }

            # 后期

            tree = fs.tree()

            print(tree)  # 输出整个目录结构

            # 使用 for in 遍历

            for node in tree.children.items():

                pass

            pass

        except Exception as error:

            print(error)

            pass
    """

    pass

