import io
import os
import json
import time
import datetime
from PIL import Image
from pymongo import MongoClient
from mongoengine import (
    connect,
    IntField,
    StringField,
    FloatField,
    DateTimeField,
    ListField,
    DictField,
    EmbeddedDocumentListField,
    EmbeddedDocument,
    Document,
    BinaryField,
)


class TileSchema(Document):
    """
    tile schema - 存储瓦片，表名由`tiles_` 和`tileset_id`组合
    """
    zoom_level  = IntField()    # zoom
    tile_column = IntField()    # x
    tile_row    = IntField()    # y
    tile_date   = IntField()    # 日期换算成时间戳，四舍五入采用int整秒表示，如果数据未提供日期信息，则默认为计算机时间纪元0点，即UTC的1970年1月1日，此时该值为0。
    tile_data   = BinaryField() # 瓦片的二进制数据


class MongoTileStorageEmbDoc(EmbeddedDocument):
    """
    MongoTileStorage, 作为 Embedded Document 嵌入到SimpleTileSetEmbDoc中
    """
    storage_type = StringField(required=True, default="MongoDB_Doc")
    path         = StringField()


class SimpleTileSetEmbDoc(EmbeddedDocument):
    """
    tileset简易对象, 作为 Embedded Document 嵌入到TileSetSchema中
    """
    min_zoom     = IntField(default=0)
    max_zoom     = IntField(default=22)
    bounds       = ListField(FloatField(required=True, default=[-180, -90, 180, 90]))  # bbox
    center       = ListField(FloatField(required=True))  # e.g. [76.275329586789, 139.153492567373, 10]
    tile_storage = EmbeddedDocumentListField(MongoTileStorageEmbDoc)


class TileSetSchema(Document):
    """
    tileset schema: 多时相瓦片集文档对象
    """
    tile_set_id = StringField(unique=True)
    owner       = StringField()
    tilejson    = StringField(default="2.1.x")
    tile_grid   = StringField(default="")
    name        = StringField(required=True)
    description = StringField()
    version     = StringField(default="1.0.0")
    attribution = StringField()
    scheme      = StringField(required=True, default="xyz")
    tiles       = ListField(StringField(required=True, default=[]))  # 瓦片URL数组，通过该地址应能够访问到瓦片,形如"http://mongo_tile_server/{tileset_id}/{z}/{x}/{y}.jpg"
    min_zoom    = IntField(default=0)
    max_zoom    = IntField(default=25)
    bounds      = ListField(FloatField(required=True))  # bbox, [-180, -90, 180, 90]
    center      = ListField(FloatField(required=True))  # e.g. [76.275329586789, 139.153492567373, 10]
    # 时相支持
    msmt_info = DictField()  # //以日期字符串为键，tileset简易对象SimpleTileSetEmbDoc为值
    date_modified = DateTimeField(default=datetime.datetime.utcnow)

    meta = \
        {
            'indexes': ['tile_set_id'],  # 索引
            'ordering': ['-date_modified'],  # -按时间倒序
            "collection": 'tile_set_schema',
        }


class StorageMongo:
    def __init__(self, db_host, db_port, db_name, tileset, tiles_path, min_zoom, max_zoom, merging):
        self.host    = db_host
        self.port    = db_port
        self.db_name = db_name
        self.merging = merging

        self.tileset_id          = tileset.get('id', '')
        self.tileset_name        = tileset.get('name', '')
        self.tileset_owner       = tileset.get('owner', '')
        self.tileset_tile_grid   = tileset.get('tile_grid', '')
        self.tileset_attribution = tileset.get('attribution', '')
        self.tileset_description = tileset.get('description', '')

        self.tiles_path = tiles_path
        self.min_zoom   = min_zoom
        self.max_zoom   = max_zoom

        db_client       = MongoClient(self.host, self.port)
        self.db_client  = db_client[self.db_name]

        self.db_engine_client = connect \
                (
                db=self.db_name,
                host=self.host,
                port=self.port,
                alias="default",
                authentication_source='admin'
            )

    def storage(self):

        msmt_info = {}

        # 集合名/表名
        tile_collection_name = 'tiles_{}'.format(self.tileset_id)

        # 瓦片
        current_collection_db = self.db_client[tile_collection_name]
        tileset_bbox = []

        # 检测瓦片目录下的不同日期

        abs_tiles_path = self.tiles_path

        # 检测瓦片目录下的不同日期
        for tile_date in os.listdir(abs_tiles_path):

            date_dir_path = os.path.join(abs_tiles_path, tile_date)
            tiles_dir_path = os.path.join(date_dir_path, 'tiles')

            # 如果不是文件夹则不继续处理
            if not os.path.isdir(date_dir_path):

                continue

            # 根据meta.json获取bbox
            meta_json_path = os.path.join(date_dir_path, 'meta.json')
            with open(meta_json_path, 'r') as fb:
                data = json.load(fb)
                bbox = data['bbox']

            if tileset_bbox == []:
                tileset_bbox = bbox

            else:
                _min_lon = min(tileset_bbox[0], bbox[0])
                _min_lat = min(tileset_bbox[1], bbox[1])
                _max_lon = max(tileset_bbox[2], bbox[2])
                _max_lat = max(tileset_bbox[3], bbox[3])

                tileset_bbox = [_min_lon, _min_lat, _max_lon, _max_lat]

                pass

            # 对应瓦片集信息中的时相字段
            ts = MongoTileStorageEmbDoc(path="mongodb://{host}:{port}/{db}/{tile_set_id}/{tile_date}".format(
                host=self.host,
                port=self.port,
                db=self.db_name,
                tile_set_id=self.tileset_id,
                tile_date=tile_date
            ))
            a_tile_set = SimpleTileSetEmbDoc(tile_storage=ts)

            msmt_info.update({
                tile_date: a_tile_set
            })

            for tile in os.listdir(tiles_dir_path):
                name, ext = os.path.splitext(tile)
                # ext = ext.strip('.').lower()

                # 获取当前瓦片的x、y、z
                z, y, x = name.split('_')
                tile_path = os.path.join(tiles_dir_path, tile)

                # 将瓦片转换成二进制
                # byte_image = io.BytesIO()
                # img = Image.open(tile_path)
                #
                # img.save(byte_image, format=ext)
                # byte_img = byte_image.getvalue()

                with open(tile_path, 'rb') as fb:
                    data_byte = fb.read()

                # 判断瓦片目录是否是时间格式,并转换成时间戳并取整
                try:
                    # 日期换算成时间戳，四舍五入采用int整秒表示，如果数据未提供日期信息，则默认为计算机时间纪元0点，即UTC的1970年1月1日，此时该值为0。
                    tmp_date = time.strptime(tile_date, "%Y-%m-%d")
                    time_stamp = int(time.mktime(tmp_date))

                except Exception as err:

                    time_stamp = 0

                # tile schema相关信息
                tile_schema = {
                    'zoom_level': int(z),
                    'tile_column': int(x),
                    'tile_row': int(y),
                    'tile_date': time_stamp,
                    'tile_data': data_byte
                }

                tile_one = current_collection_db.find_one({'zoom_level': int(z), 'tile_column': int(x), 'tile_row': int(y), 'tile_date': time_stamp, })

                if tile_one is None:
                    # 插入mongodb数据库中
                    current_collection_db.insert_one(tile_schema)

                    continue

                tile_data_byte_mongo = tile_one.get('tile_data')

                if self.merging == 0:
                    # 仅保留原瓦片
                    continue

                elif self.merging == 1:
                    # 替换原瓦片
                    tile_byte = data_byte
                    pass

                elif self.merging == 2:
                    # 合并同名瓦片（原瓦片在上）
                    new_img = Image.open(io.BytesIO(data_byte))
                    origin_img = Image.open(io.BytesIO(tile_data_byte_mongo))

                    new_img.paste(origin_img, (0, 0), origin_img)

                    image_arr_byte = io.BytesIO()
                    new_img.save(image_arr_byte, "PNG")

                    tile_byte = image_arr_byte.getvalue()

                    pass

                elif self.merging == 3:
                    # 合并同名瓦片（原瓦片在下）
                    new_img = Image.open(io.BytesIO(data_byte))
                    origin_img = Image.open(io.BytesIO(tile_data_byte_mongo))

                    origin_img.paste(new_img, (0, 0), new_img)

                    image_arr_byte = io.BytesIO()
                    origin_img.save(image_arr_byte, "PNG")

                    tile_byte = image_arr_byte.getvalue()

                    pass

                current_collection_db.find_one_and_update(
                    {
                        'zoom_level': int(z),
                        'tile_column': int(x),
                        'tile_row': int(y),
                        'tile_date': time_stamp,
                    },
                    {
                        "$set": {
                            'tile_data': tile_byte,
                        }
                    },
                    upsert=True
                )

        tileset_center = [
            (tileset_bbox[0] + tileset_bbox[2]) / 2,
            (tileset_bbox[1] + tileset_bbox[3]) / 2,
            self.min_zoom
        ]

        try:
            # 保存瓦片集信息到mongodb中
            tile_set = TileSetSchema \
                    (
                    tile_set_id=self.tileset_id,
                    owner=self.tileset_owner,
                    tilejson='2.1.x',
                    tile_grid=self.tileset_tile_grid,
                    name=self.tileset_name,
                    description=self.tileset_description,
                    version='1.0.0',
                    attribution=self.tileset_attribution,
                    scheme='xyz',
                    min_zoom=self.min_zoom,
                    max_zoom=self.max_zoom,
                    bounds=tileset_bbox,
                    center=tileset_center,
                    msmt_info=msmt_info,
                )

            tile_set.save()

        except Exception as err:
            # self.tileset_id瓦片集标识可能会重复
            pass

        pass
