import io
import json
import os
import shutil
import traceback
from multiprocessing import Process

from PIL import Image
from map_lib.mtp.mtp_writer import MtpWriter
from map_lib.mtp.mtp_reader import MtpReader
from map_lib.storage.file import FileStorage
from map_lib.mtp.mtp_modifier import MtpModifier
from map_lib.mtp.mtp_batch import MtpBatch
from starearth.utils import private_key, dev_key


def create_ptp(ptp_path, map_name='satellite', map_id='r_google', tile_ext='jpg', tile_size=256, mtp_id='PTP'):
    """创建ptp包"""

    fs = FileStorage()
    fs.create(ptp_path)
    tile_ranges = {}
    ptp_name = os.path.basename(ptp_path).split('.')[0]
    min_zoom_level, max_zoom_level, base_zoom_level, base_x, base_y = ptp_name.split('-')

    min_zoom_level, max_zoom_level, base_zoom_level, base_x, base_y = \
        int(min_zoom_level), int(max_zoom_level), int(base_zoom_level), int(base_x), int(base_y)

    for zoom_level in range(min_zoom_level, max_zoom_level + 1):
        start_x = pow(2, zoom_level - base_zoom_level) * base_x
        start_y = pow(2, zoom_level - base_zoom_level) * base_y

        tile_ranges[zoom_level] = [start_x, start_x + pow(2, zoom_level - base_zoom_level) - 1,
                                   start_y, start_y + pow(2, zoom_level - base_zoom_level) - 1]

    ranges = {str(k): tile_ranges[k] for k in sorted(tile_ranges.keys())}

    metadata = {
        'map_name': map_name,
        'packet_name': ptp_name,
        'map_id': map_id,
        'min_zoom_level': min_zoom_level,
        'max_zoom_level': max_zoom_level,
        'tile_file_ext': tile_ext,
        'bounds': (-180, -90, 180, 90)
    }

    mtp_writer = MtpWriter(fs,  # FileStorage 句柄
                           private_key,  # 私有key
                           dev_key,  # 设备key
                           metadata,  # mtp 包的元数据
                           ranges,  # 包含的瓦片层，创建后不能修改
                           tile_size,  # 瓦片尺寸，该实例为 256 x 256
                           mtp_id)  # mtp 包的ID，例如：PTP、DTP
    return mtp_writer


class StoragePTP:
    def __init__(self, ptp_path, tiles_path):

        self.ptp_path = ptp_path  # ptp的临时目录
        self.tiles_path = tiles_path  # 瓦片的临时目录

        pass

    def get_ptp_name(self, x, y, z, base_list):

        for base in base_list:
            minz, maxz, basez = base[0], base[1], base[2]
            if minz <= z <= maxz:
                return "{minz}-{maxz}-{basez}-{basex}-{basey}".format(minz=minz, maxz=maxz, basez=basez,
                                                                      basex=x // 2 ** (z - basez),
                                                                      basey=y // 2 ** (z - basez))
        raise Exception('zoom out of range base_list.')

    def save(self, package_file_dir, tile_dir, tile_list, package_rule, tile_format, tile_size):
        """瓦片存储到ptp文件中"""

        tiles_path = os.path.join(tile_dir, 'tiles')
        for x, y, z in tile_list:  # 瓦片对象
            package_file_name = self.get_ptp_name(x, y, z, package_rule)
            package_file_path = str(os.path.join(str(package_file_dir), package_file_name + '.ptp'))
            tile_file_name = '{}_{}_{}.{}'.format(z, y, x, tile_format)

            tile_file = os.path.join(tiles_path, tile_file_name)

            try:
                with open(tile_file, 'rb') as f:
                    tile_content = f.read()
            except Exception as err:
                traceback.print_exc()

                continue

            if os.path.exists(package_file_path):
                fs = FileStorage()
                fs.open(package_file_path, "rb+")
                mtp_writer = MtpModifier(fs, private_key, dev_key)

            else:
                if tile_format == 'jpeg':

                    mtp_writer = create_ptp(package_file_path)

                elif tile_format == 'png':

                    mtp_writer = create_ptp(package_file_path, tile_ext='png', tile_size=tile_size)

                elif tile_format == 'bil':

                    mtp_writer = create_ptp(ptp_path=package_file_path, map_name='dem', map_id='r_google',
                                                 tile_ext='bil', tile_size=128, mtp_id='PTP')

                else:

                    mtp_writer = create_ptp(package_file_path)

            mtp_writer.add_tile_with_data(tile_content, z, x, y)

            mtp_writer.close()

        if os.path.exists(os.path.join(tile_dir, 'imagesheet.geojsonl')):
            shutil.move(os.path.join(tile_dir, 'imagesheet.geojsonl'),
                        os.path.join(package_file_dir, 'imagesheet.geojsonl'))

        if os.path.exists(os.path.join(tile_dir, 'meta.json')):

            shutil.move(os.path.join(tile_dir, 'meta.json'), os.path.join(package_file_dir, 'meta.json'))

    def storage(self, package_rule, tile_format, tile_size):

        p_list = []

        for date in os.listdir(str(self.tiles_path)):

            _tile_dir = os.path.join(str(self.tiles_path), date)
            _package_file_dir = os.path.join(self.ptp_path, date)

            if not os.path.exists(_package_file_dir):
                os.makedirs(_package_file_dir, exist_ok=True)

            with open(os.path.join(_tile_dir, 'tiles_list.json'), 'r') as f:

                _tile_list = json.loads(f.read())
                p_list.append(Process(target=self.save,
                                      args=(_package_file_dir, _tile_dir, _tile_list, package_rule, tile_format, tile_size)))
        for p in p_list:
            p.start()

        for p in p_list:
            p.join()

        for p in p_list:
            if p.exitcode != 0:
                raise Exception('多进程打包执行失败.')


def ptp_merge(output_path, storage_dic, tileset, package_name, package_path_list):
    # 遍历package_path_list有则合并没则拷贝
    for package_path in package_path_list:
        date = package_path.split('/')[-1]

        outpath_ptp = os.path.join(output_path, date, package_name + '.ptp')
        new_ptp = os.path.join(package_path, package_name + '.ptp')
        # 合包
        if os.path.exists(outpath_ptp):

            # 该合包方式无法处理半个瓦片的情况，故不使用
            # MtpBatch.merge(os.path.join(package_path, package_name + '.ptp'),
            #                os.path.join(output_path, date, package_name + '.ptp'), private_key, dev_key)

            if not os.path.exists(new_ptp):
                continue

            output_ptp_bak = outpath_ptp + ".bak"
            shutil.move(outpath_ptp, output_ptp_bak)

            # 以下合包方式参照旧版slice
            fs1 = FileStorage()
            fs1.open(output_ptp_bak)
            mtp_reader1 = MtpReader(fs1, private_key, dev_key)

            fs2 = FileStorage()
            fs2.open(new_ptp)
            mtp_reader2 = MtpReader(fs2, private_key, dev_key)

            metadata = mtp_reader1.metadata
            map_name = metadata.get('map_name')
            map_id = metadata.get('map_id')
            tile_ext = metadata.get('tile_file_ext')

            mtp_writer = create_ptp(outpath_ptp, map_name=map_name, map_id=map_id, tile_ext=tile_ext)

            for level, tile_range in mtp_reader2.tile_ranges.items():
                minx, maxx, miny, maxy, _ = tile_range
                for x in range(minx, maxx + 1):
                    for y in range(miny, maxy + 1):

                        buff1 = mtp_reader1.get_tile(level, x, y)
                        buff2 = mtp_reader2.get_tile(level, x, y)

                        if not buff1 and not buff2:
                            continue

                        if not buff1 and buff2:
                            mtp_writer.add_tile_with_data(buff2, level, x, y)

                        if not buff2 and buff1:
                            mtp_writer.add_tile_with_data(buff1, level, x, y)

                        if buff2 and buff1:

                            bytes = io.BytesIO(buff1)
                            img1 = Image.open(bytes)

                            bytes2 = io.BytesIO(buff2)
                            img2 = Image.open(bytes2)

                            img1.paste(img2, (0, 0), img2)

                            output_io = io.BytesIO()
                            img1.save(output_io, format=img1.format)

                            mtp_writer.add_tile_with_data(output_io.getvalue(), level, x, y)

                            del img1
                            del img2

            del mtp_reader1
            del mtp_reader2
            del mtp_writer

            if os.path.exists(output_ptp_bak):
                os.remove(output_ptp_bak)

        else:

            if not os.path.exists(os.path.join(output_path, date)):
                # 创建路径 输出约定路径 + output_dir + 日期
                os.makedirs(os.path.join(output_path, date), exist_ok=True)

            shutil.move(os.path.join(package_path, package_name + '.ptp'), os.path.join(output_path, date, package_name + '.ptp'))
