import datetime
import hashlib
import json
import os
import shutil
import sqlite3
import mercantile
from io import BytesIO
from multiprocessing import Process
from PIL import Image


def optimize_connection(cur):
    cur.execute("""PRAGMA synchronous=0""")
    cur.execute("""PRAGMA locking_mode=EXCLUSIVE""")
    cur.execute("""PRAGMA journal_mode=DELETE""")


def sqlite_connect(package_file):
    try:
        return sqlite3.connect(package_file)
    except Exception as e:
        print(e)
        return None


def optimize_database(cur):
    cur.execute("""ANALYZE;""")

    cur.isolation_level = None
    cur.execute("""VACUUM;""")
    cur.isolation_level = ''


def smtiles_setup(cur):
    # 创建表
    cur.execute(
        """
        CREATE TABLE IF NOT EXISTS map (
           zoom_level INTEGER,
           tile_column INTEGER,
           tile_row INTEGER,
           tile_id TEXT,
           resolution TEXT,
           create_time TEXT 
        );
        """
    )
    cur.execute(
        """
        CREATE TABLE IF NOT EXISTS images (
            tile_data BLOB,
            tile_id TEXT
        );
        """
    )
    cur.execute(
        """
        CREATE TABLE IF NOT EXISTS metadata (
            name TEXT,
            value TEXT

        );
        """
    )

    # 创建索引
    cur.execute(
        """
        CREATE UNIQUE INDEX IF NOT EXISTS map_index ON map (zoom_level, tile_column, tile_row, resolution);
        """
    )
    cur.execute(
        """
        CREATE UNIQUE INDEX IF NOT EXISTS images_id ON images (tile_id);
        """
    )
    cur.execute(
        """
        CREATE UNIQUE INDEX IF NOT EXISTS name ON metadata (name);
        """
    )

    # 创建视图
    cur.execute(
        """
        CREATE VIEW IF NOT EXISTS tiles AS
            SELECT
                map.zoom_level AS zoom_level,
                map.tile_column AS tile_column,
                map.tile_row AS tile_row,
                map.resolution AS resolution, 
                images.tile_data AS tile_data
            FROM map JOIN images ON images.tile_id = map.tile_id;
        """
    )


class StorageSmtiles:
    def __init__(self, package_path, tiles_path):
        self.package_path = package_path  # ptp的临时目录
        self.tiles_path = tiles_path  # 瓦片的临时目录

        pass

    def calc_package_name(self, x, y, z, base_list):
        for base in base_list:
            minz, maxz, basez = base[0], base[1], base[2]
            if minz <= z <= maxz:
                return "{minz}-{maxz}-{basez}-{basex}-{basey}".format(minz=minz, maxz=maxz, basez=basez,
                                                                      basex=x // 2 ** (z - basez),
                                                                      basey=y // 2 ** (z - basez))
        raise Exception('zoom out of range base_list.')

    def save(self, package_file_dir, tile_dir, tile_list, package_rule, tile_format, param):

        for x, y, z in tile_list:  # 瓦片对象
            package_file_name = self.calc_package_name(x, y, z, package_rule)
            min_z, max_z, base_z, pack_x, pack_y = map(lambda _x: int(_x), package_file_name.split('-'))

            package_file_path = str(os.path.join(str(package_file_dir), package_file_name + '.smtiles'))

            tile_file_name = '{}_{}_{}.{}'.format(z, y, x, tile_format)
            tile_file = os.path.join(tile_dir, 'tiles', tile_file_name)

            con = sqlite_connect(package_file_path)
            if not con:
                raise Exception(u'数据库连接失败:%s' % package_file_path)
            cur = con.cursor()
            optimize_connection(cur)

            smtiles_setup(cur)

            with open(tile_file, 'rb') as f:
                tile_content = f.read()
            # 根据内容计算tile_id, 相同的瓦片会丢掉,全透明或全白
            tile_id = hashlib.md5(tile_content).hexdigest()
            tile_id = tile_id.split(r'  ')[0]

            # 写入images
            cur.execute(
                """INSERT OR IGNORE INTO images (tile_data, tile_id) VALUES (?, ?); """,
                (sqlite3.Binary(tile_content), tile_id))

            # 写入map
            cur.execute(
                """INSERT OR IGNORE INTO map (zoom_level, tile_column, tile_row, tile_id, create_time) VALUES (?, ?, ?, ?, ?);""",
                (z, x, y, tile_id, datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")))

            # 写metadata
            bounds = mercantile.bounds(pack_x, pack_y, base_z)

            metadata = {
                'name': package_file_name,
                'type': 'baselayer',
                'version': '1.1',
                'description': param.get('description', ''),
                'format': tile_format,
                'bounds': ",".join([str(i) for i in bounds]),
                'ext_spec_version': '201310',
                'axis_origin': '-180.0,90.0' if param.get(
                    'tile_grid') == 'WGS1984Quad' else '-20037508.3427892,20037508.3427892',
                # 经纬度-180.0,90.0  墨卡托 [-20037508.3427892,20037508.3427892]
                'axis_positive_direction': 'RightDown',
                'crs_wkid': str(param.get('epsg')),
                'crs_wkt': str(param.get('wkt')),
                'tile_height': str(param.get('tile_size')),
                'tile_width': str(param.get('tile_size')),
                'resolutions': str(param.get('resolutions')),
                'scales': '',
                'transparent': 'false',
                'mapStatusHashCode': '',
                'map_parameter': '',
                'compatible': 'false' if param.get('tile_grid') == 'WGS1984Quad' else 'true'
                # compatible 经纬度false 墨卡托true
            }

            for name, value in metadata.items():
                cur.execute('REPLACE INTO metadata (name,value) VALUES (?,?)', (name, value))

            con.commit()
            optimize_database(con)
            con.close()

        if os.path.exists(os.path.join(tile_dir, 'imagesheet.geojsonl')):
            shutil.move(os.path.join(tile_dir, 'imagesheet.geojsonl'),
                        os.path.join(package_file_dir, 'imagesheet.geojsonl'))

        if os.path.exists(os.path.join(tile_dir, 'meta.json')):
            shutil.move(os.path.join(tile_dir, 'meta.json'), os.path.join(package_file_dir, 'meta.json'))

    def storage(self, package_rule, tile_format, param):

        p_list = []

        for date in os.listdir(str(self.tiles_path)):

            _tile_dir = os.path.join(str(self.tiles_path), date)
            _package_file_dir = os.path.join(self.package_path, date)

            if not os.path.exists(_package_file_dir):
                os.makedirs(_package_file_dir, exist_ok=True)

            with open(os.path.join(_tile_dir, 'tiles_list.json'), 'r') as f:
                _tile_list = json.loads(f.read())
                p_list.append(Process(target=self.save,
                                      args=(_package_file_dir, _tile_dir, _tile_list, package_rule, tile_format, param)))

        for p in p_list:
            p.start()

        for p in p_list:
            p.join()

        for p in p_list:
            if p.exitcode != 0:
                raise Exception('多进程打包执行失败.')


def smtiles_merge(output_path, storage_dic, tileset, package_name, package_path_list):

    # 遍历package_path_list有则合并没则拷贝
    for package_path in package_path_list:
        date = package_path.split('/')[-1]
        # 合包
        if os.path.exists(os.path.join(output_path, date, package_name + '.smtiles')):
            con1 = sqlite_connect(os.path.join(package_path, package_name + '.smtiles'))
            con2 = sqlite_connect(os.path.join(output_path, date, package_name + '.smtiles'))
            if not con1:
                raise Exception(u'数据库连接失败:%s' % os.path.join(package_path, package_name + '.smtiles'))
            if not con2:
                raise Exception(u'数据库连接失败:%s' % os.path.join(output_path, date, package_name + '.smtiles'))

            cur1 = con1.cursor()
            cur2 = con2.cursor()
            optimize_connection(cur1)
            optimize_connection(cur2)

            cur1.execute("""SELECT * FROM map;""")
            tiles = cur1.fetchall()

            for index, tile in enumerate(tiles):
                z, x, y = tile[0], tile[1], tile[2]

                # 写入瓦片
                cur1.execute(
                    """SELECT tile_data FROM tiles WHERE zoom_level=%s AND tile_column=%s AND tile_row=%s;""" % (
                        z, x, y))
                result1 = cur1.fetchone()
                if result1:
                    tile_content1 = result1[0]
                else:
                    continue

                cur2.execute(
                    """SELECT tile_data FROM tiles WHERE zoom_level=%s AND tile_column=%s AND tile_row=%s;""" % (
                        z, x, y))
                result2 = cur2.fetchone()
                if result2:
                    tile_content2 = result2[0]
                    # PIL
                    img1 = Image.open(BytesIO(tile_content1))
                    img2 = Image.open(BytesIO(tile_content2))
                    mask = img1.getchannel('A')
                    img = Image.composite(img1, img2, mask)
                    tile_content = img.tobytes()

                    cur2.execute(
                        """SELECT tile_id FROM map WHERE zoom_level=? AND tile_column=? AND tile_row=?""",
                        (z, x, y))
                    tile_id = cur2.fetchone()[0]

                    new_tile_id = hashlib.md5(tile_content).hexdigest()
                    new_tile_id = new_tile_id.split(r'  ')[0]
                    # 写入images
                    # print(z, x, y, tile_id)
                    cur2.execute("""UPDATE OR IGNORE images SET tile_data=?,tile_id=? WHERE tile_id=?;""",
                                 (sqlite3.Binary(tile_content), new_tile_id, tile_id))
                    # 写入map
                    cur2.execute(
                        """UPDATE OR IGNORE map SET tile_id=? WHERE tile_id=?;""", (new_tile_id, tile_id))
                else:
                    # 直接写入
                    tile_content = tile_content1
                    tile_id = tile[3]
                    cur2.execute(
                        """INSERT OR IGNORE INTO images (tile_data, tile_id) VALUES (?, ?);""",
                        (sqlite3.Binary(tile_content), tile_id))
                    cur2.execute(
                        """INSERT OR IGNORE INTO map (zoom_level, tile_column, tile_row, tile_id) VALUES (?,?,?,?);""",
                        (z, x, y, tile_id))

            con2.commit()
            optimize_database(con2)
            con1.close()
            con2.close()

        else:

            if not os.path.exists(os.path.join(output_path, date)):
                # 创建路径 输出约定路径 + output_dir+ 日期
                os.makedirs(os.path.join(output_path, date), exist_ok=True)

            shutil.move(os.path.join(package_path, package_name + '.smtiles'),
                        os.path.join(output_path, date, package_name + '.smtiles'))
