import os
import json
from jsonschema import validate, ValidationError
from app.utils.json_utils import JSONDict
from app.defines import TileSetType

class TileSet(object):
    """
    对 `tileset.json`进行建模，使用jsonschema库提供读写验证
    """
    Schema = \
        {
            "type": "object",
            "properties":
                {
                    "id": {"type": "string"},
                    "name": {"type": "string"},
                    "description": {"type": "string"},
                    "scheme": {"type": "string"},  # "xyz",
                    "tilejson": {"type": "string"},  # "2.1.x",
                    "tile_grid": {"type": "string"},  # WebMercatorQuad、WGS1984Quad、GoogleEarth
                    "type": {"type": "string"},
                    "version": {"type": "string"},
                    "minzoom": {"type": "number"},
                    "maxzoom": {"type": "number"},
                    "bounds":
                        {
                            "type": "array",
                            "items": {"type": "number", "maxItems": 4}
                        },
                    "center":
                        {
                            "type": "array",
                            "items": {"type": "number", "maxItems": 3}
                        },
                    "tiles":
                        {
                            "type": "array",
                            "items": {"type": "string"}
                        },
                    "owner": {"type": "string"},
                    "attribution": {"type": "string"},
                    "msmt_info":
                        {
                            "type": "array",
                            "items": {"type": "object"}
                        }
                },
            "required": ["id", "scheme", "tilejson", "maxzoom", "minzoom"]
        }

    def __init__(self):
        self.id          = None
        self.name        = None
        self.description = None
        self.scheme      = None
        self.tilejson    = None
        self.tile_grid   = None
        self.type        = None
        self.version     = None

        self.minzoom     = None
        self.maxzoom     = None
        self.bounds      = None
        self.center      = None
        self.tiles       = None

        self.owner       = None
        self.attribution = None

        self.msmt_info   = None

        pass

    def load(self, path):
        """
        从路径加载tileset.json,并构造Tileset对象
        :param path:
        """
        with open(TileSet._resolve(path), 'r') as fp:

            data = json.load(fp)
            try:
                validate(data, TileSet.Schema)
            except ValidationError as e:
                print(e, "please check your tileset.json file.")

            self.id        = data['id'      ]
            self.scheme    = data['scheme'  ]
            self.tilejson  = data['tilejson']
            self.tile_grid = data['tile_grid']

            self.name        = data.get("name")
            self.description = data.get("description")

            self.type    = data.get("type")
            self.version = data.get("description", "1.0.0")

            self.minzoom = data['minzoom'    ]
            self.maxzoom = data['maxzoom'    ]
            self.bounds  = data.get("bounds")
            self.center  = data.get("center")
            self.tiles   = data.get("tiles", [])

            self.owner       = data.get("owner", "mapscloud.cn")
            self.attribution = data.get("attribution", "mapscloud.cn")

            try:
                _msmt_info = data.get("msmt_info")
                for item in _msmt_info:
                    TileSet.msmt_item_validate(item)
                self.msmt_info = _msmt_info

            except ValidationError as e:
                print(e, "please check your t items in tileset.json file.")

            pass

        pass

    def save(self, path):
        """
        给定地址,TileSet 对象写入文件 'tileset.json'
        :param path:
        """
        with open(TileSet._resolve(path), 'w') as fp:
            data = JSONDict()

            data.id = self.id
            data.scheme = self.scheme or "xyz"
            data.tilejson  = self.tilejson or "2.1.x"
            data.tile_grid = self.tile_grid or ""
            data.name = self.name or ""
            data.description = self.description or ""
            data.type = self.type or TileSetType.IMAGE_TILE
            data.version = self.version or "1.0.0"

            data.minzoom = self.minzoom or 0
            data.maxzoom = self.maxzoom or 25
            data.bounds  = self.bounds
            data.center  = self.center
            data.tiles   = self.tiles or []
            data.owner   = self.owner or "mapscloud.cn"

            data.attribution = self.attribution or "mapscloud.cn"
            data.msmt_info   = self.msmt_info or []

            try:
                validate(data, TileSet.Schema)
            except ValidationError as e:
                print(e, "TileSet validate error.")

            json.dump(data, fp, ensure_ascii=False)

            pass

        pass

    @staticmethod
    def construct_from(path):
        """
        静态方法: 从路径加载tileset.json,并构造Tileset对象
        :param path:
        :return:
        """
        instance = TileSet()
        instance.load(path)

        return instance

    @staticmethod
    def _resolve(path):
        if os.path.isfile(path):
            path = os.path.dirname(path)

            pass

        return os.path.join(path, 'tileset.json')

    @staticmethod
    def msmt_item_validate(json_data):
        """
        验证每个时相及其存储
        :param json_data:
        """
        msmt_item_schema = {
            "type": "object",
            "patternProperties": {"\d{4}-\d{1,2}-\d{1,2}": {  # 键是变化的，使用patternProperties的正则表达式描述key
                "type": "object",
                "properties": {
                    "storage_type": {"type": "string"},
                    "path": {"type": "string"},
                },
                "required": ["storage_type", "path"]
            }
            }
        }

        validate(json_data, msmt_item_schema)



