import rasterio
import numpy as np
from osgeo import gdal, osr


def create_overviews(input_file, block_size=512):
    """
    创建金字塔：
    1. 如果输入文件已有内置或外置金字塔，返回overview个数；
    2. 如果输入文件无金字塔，则创建金字塔，再返回overview个数。
    """
    ds = gdal.Open(input_file, gdal.GA_ReadOnly)
    band1 = ds.GetRasterBand(1)
    ovr_count = band1.GetOverviewCount()
    if ovr_count != 0:
        return ovr_count
    width = ds.RasterXSize
    height = ds.RasterYSize

    n_level = 0
    overview = 1

    while min(width // overview, height // overview) > block_size:
        overview *= 2
        n_level += 1

    overview_list = [2 ** j for j in range(1, n_level + 1)]
    ds.BuildOverviews(overviewlist=overview_list)
    return len(overview_list)


def getRealFourCorners(datafile):
    """获取角点坐标"""
    overviews = create_overviews(datafile)
    # print('overviews:', overviews)
    with rasterio.open(datafile, overview_level=overviews - 1, mode='r') as src:
        images = src.read(1)
        width = src.width
        height = src.height
        value_arr = np.nonzero(images[0])
        if value_arr[0].size > 0:
            aff = src.transform
            # print(src.bounds)
            for i in range(0, height, 1):
                value_arr = np.nonzero(images[i])
                if value_arr[0].size > 0:
                    ul_x, ul_y = aff * (value_arr[0][0], i)
                    row_top = (ul_x, ul_y)
                    # print ('hang----- top: ', ul_x, ul_y)
                    break

            for i in range(height - 1, 0, -1):
                value_arr = np.nonzero(images[i])
                if value_arr[0].size > 0:
                    ul_x, ul_y = aff * (value_arr[0][value_arr[0].size - 1], i)
                    row_bottom = (ul_x, ul_y)
                    # print ('hang----- bottom: ', ul_x, ul_y)
                    break

            for i in range(width):
                value_arr = np.nonzero(images[:, i])
                if value_arr[0].size > 0:
                    ul_x, ul_y = aff * (i, value_arr[0][value_arr[0].size - 1])
                    column_left = (ul_x, ul_y)
                    # print('shu----- left: ', ul_x, ul_y)
                    break

            for i in range(width - 1, 0, -1):
                value_arr = np.nonzero(images[:, i])
                if value_arr[0].size > 0:
                    ul_x, ul_y = aff * (i, value_arr[0][0])
                    column_right = (ul_x, ul_y)
                    # print ('shu----- right: ', ul_x, ul_y)
                    break
            # bbox = src.bounds
            # print(bbox)
            # row_top = (bbox.left, bbox.top)
            # row_bottom = (bbox.right, bbox.bottom)
            # column_left = (bbox.left, bbox.bottom)
            # column_right = (bbox.right, bbox.top)
        else:
            with rasterio.open(datafile, overview_level=0, mode='r') as src:
                images = src.read(1)
                width = src.width
                height = src.height
                aff = src.transform
                # print(src.bounds)
                for i in range(0, height, 1):
                    value_arr = np.nonzero(images[i])
                    if value_arr[0].size > 0:
                        ul_x, ul_y = aff * (value_arr[0][0], i)
                        row_top = (ul_x, ul_y)
                        # print ('hang----- top: ', ul_x, ul_y)
                        break

                for i in range(height - 1, 0, -1):
                    value_arr = np.nonzero(images[i])
                    if value_arr[0].size > 0:
                        ul_x, ul_y = aff * (value_arr[0][value_arr[0].size - 1], i)
                        row_bottom = (ul_x, ul_y)
                        # print ('hang----- bottom: ', ul_x, ul_y)
                        break

                for i in range(width):
                    value_arr = np.nonzero(images[:, i])
                    if value_arr[0].size > 0:
                        ul_x, ul_y = aff * (i, value_arr[0][value_arr[0].size - 1])
                        column_left = (ul_x, ul_y)
                        # print('shu----- left: ', ul_x, ul_y)
                        break

                for i in range(width - 1, 0, -1):
                    value_arr = np.nonzero(images[:, i])
                    if value_arr[0].size > 0:
                        ul_x, ul_y = aff * (i, value_arr[0][0])
                        column_right = (ul_x, ul_y)
                        # print ('shu----- right: ', ul_x, ul_y)
                        break
    return row_top, row_bottom, column_left, column_right


def transform_point(point, from_srs_id, to_srs_id, reverse=False):
    """经纬度、墨卡托相互转换"""
    mer = osr.SpatialReference()
    mer.ImportFromEPSG(from_srs_id)

    geo = osr.SpatialReference()
    geo.ImportFromEPSG(to_srs_id)

    if not reverse:
        t = osr.CoordinateTransformation(mer, geo)
    else:
        t = osr.CoordinateTransformation(geo, mer)

    return t.TransformPoint(*point)


def get_dem_zoom(pixelSize, dem_tile_size ,MAX_ZOOM_LEVEL=32):
    RES_FACT = 360.0 / dem_tile_size

    for i in range(1, MAX_ZOOM_LEVEL):
        res = RES_FACT / 2 ** i
        pre_res = RES_FACT / 2 ** (i - 1)
        if res < pixelSize < pre_res:
            if pixelSize <= (res + pre_res) / 2:  # 返回更接近的级别
                return i
            else:
                return i - 1
    return MAX_ZOOM_LEVEL - 1
